
#include <byteswap.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <libpq-fe.h>

typedef struct pgbench_account
{
	int aid;
	int bid;
	int abalance;
	char filler[85];
} pgbench_account;

typedef struct load_cursor
{
	pgbench_account **accounts;
	int cursor;
} load_cursor;


static const char* pghost;
static const char* pgport;
static const char* dbname;
static const char* user;
static const char* password;

static int process_callback(PGresult *res, const PGdataValue*, int col_count, void *user_data);
static void process_single_row_mode(PGresult *res, pgbench_account **list, int col_count, int cursor);


static int
process_callback(PGresult *res, const PGdataValue* vals,
				 int col_count, void *user_data)
{
	pgbench_account *acct;
	load_cursor *cursor = user_data;
	int i;

	acct = malloc(sizeof(struct pgbench_account));
	cursor->accounts[cursor->cursor] = acct;

	cursor->cursor++;

	for (i = 0; i < col_count; i++)
	{
		/*
		 * Skip over null values
		 */

		if (vals[i].len < 0)
			continue;

		switch (i)
		{
		case 0:
			acct->aid = *((int*)vals[i].value);
			acct->aid = __bswap_32(acct->aid);
			break;

		case 1:
			acct->bid = *((int*)vals[i].value);
			acct->bid = __bswap_32(acct->bid);
			break;

		case 2:
			acct->abalance = *((int*)vals[i].value);
			acct->abalance = __bswap_32(acct->abalance);
			break;

		case 3:
			memcpy(&acct->filler[0], vals[i].value, vals[i].len);
			break;
		}
	}

	return 1;
}


static void 
process_single_row_mode(PGresult *res, pgbench_account **list, int col_count, int cursor)
{
	pgbench_account *acct;
	int i;

	acct = malloc(sizeof(struct pgbench_account));
	list[cursor] = acct;

	for (i = 0; i < col_count; i++)
	{
		/*
		 * Skip over null values
		 */

		if (PQgetisnull(res, 0, i))
			continue;

		switch (i)
		{
		case 0:
			acct->aid = *((int*)PQgetvalue(res, 0, i));
			acct->aid = __bswap_32(acct->aid);
			break;

		case 1:
			acct->bid = *((int*)PQgetvalue(res, 0, i));
			acct->bid = __bswap_32(acct->bid);
			break;

		case 2:
			acct->abalance = *((int*)PQgetvalue(res, 0, i));
			acct->abalance = __bswap_32(acct->abalance);
			break;

		case 3:
			memcpy(&acct->filler[0], PQgetvalue(res, 0, i), PQgetlength(res, 0, i));
			break;
		}
	}
}

static PGconn*
doConnect(void)
{
	PGconn *conn;
	int8_t new_pass;
	const char *keywords[7];
	const char *values[7];

	keywords[0] = "hostaddr";
	values[0] = pghost;
	keywords[1] = "port";
	values[1] = pgport;
	keywords[2] = "user";
	values[2] = user;
	keywords[3] = "password";
	values[3] = password;
	keywords[4] = "dbname";
	values[4] = dbname;
	keywords[5] = 0;
	values[5] = 0;

	conn = PQconnectdbParams(keywords, values, 0);

	if (!conn)
	{
		fprintf(stderr, "connectiont to database \"%s\" failed\n",
				dbname);
		return 0;
	}

	if (PQstatus(conn) == CONNECTION_BAD)
	{
		fprintf(stderr, "connection to database \"%s\" failed:\n%s",
				dbname,
				PQerrorMessage(conn));
		PQfinish(conn);

		return 0;
	}

	return conn;
}

int main(int argc, char *argv[])
{
	static struct option long_options[] = {
		{"host", required_argument, 0, 'h'},
		{"port", required_argument, 0, 'p'},
		{"username", required_argument, 0, 'U'},
		{"password", required_argument, 0, 'P'},
		{"dbname", required_argument, 0, 'd'},
		{"mode", required_argument, 0, 'm'},
		{0, 0, 0, 0}
	};

	int ret, single_row_mode, col_count = 0, c, optidx, i;
	PGconn *conn;
	PGresult *res;
	load_cursor cursor = { };

	while ((c = getopt_long(argc, argv, "h:p:U:P:d:m:", long_options, &optidx)) != -1)
	{
		switch (c)
		{
			case 'h':
				pghost = strdup(optarg);
				break;

			case 'p':
				pgport = strdup(optarg);
				break;

			case 'U':
				user = strdup(optarg);
				break;

			case 'P':
				password = strdup(optarg);
				break;

			case 'd':
				dbname = strdup(optarg);
				break;

			case 'm':
			
				if (strcmp(optarg, "cb") == 0)
					single_row_mode = 0;
				else
					single_row_mode = 1;

				break;
		}
	}	
	
	
	/*
	 * Make the connection
	 */
	conn = doConnect();

	if (!conn)
		exit(1);

	cursor.cursor = 0;
	cursor.accounts = malloc(sizeof(pgbench_account*) * 100000);

	/*
	 * Send the query asyncronously
	 */
	ret = PQsendQueryParams(conn, 
							"SELECT * FROM pgbench_accounts LIMIT 100000;",
							0,				/* nParams */
							0,				/* paramTypes */
							0,				/* paramValues */
							0, 				/* paramLengths */
							0,				/* paramFormats */
							1);				/* use Binary format */

	if (single_row_mode)
	{
		/*
		 * Set the single row mode before consuming the input.
		 */
		ret = PQsetSingleRowMode(conn);
		PQconsumeInput(conn);

		printf("\nConsuming records using single row mode...");

		while ((res = PQgetResult(conn)))
		{
			if (col_count == 0)
				col_count = PQnfields(res);

			if (PQntuples(res))
			{
				process_single_row_mode(res, cursor.accounts, col_count, cursor.cursor);
				cursor.cursor++;
			}

			PQclear(res);
		}

		printf("\n%d rows processed", cursor.cursor);
	}
	else
	{
		/*
		 * PQconsumeInput will eventually invoke process_callback.
		 */
		ret = PQsetRowProcessor(conn, process_callback, &cursor);
		
		printf("\nConsuming records using callback...");

		PQconsumeInput(conn);

		res = PQgetResult(conn);
		PQclear(res);

		printf("\n%d rows processed", cursor.cursor);
	}

	PQfinish(conn);

	/*
	 * Clean up the records allocated in memory.
	 */
	for (i = 0; i < cursor.cursor; i++)
		free(cursor.accounts[i]);

	free(cursor.accounts);

	printf("\nTest finished\n");

	exit(0);
}

